#include <Python.h>
#include "structmember.h"

#include "expat/expat.h"

#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
# define PY_SSIZE_T_MAX INT_MAX
# define PY_SSIZE_T_MIN INT_MIN
#endif

typedef struct {
    PyObject_HEAD
    XML_Parser parser;
    PyObject*  fd;
    PyObject*  read;
    PyObject*  read_args;
    PyObject*  dict_singleton;
    PyObject*  td_singleton;
    int        buffersize;
    Py_ssize_t queue_read_idx;
    Py_ssize_t queue_write_idx;
    Py_ssize_t text_alloc;
    Py_ssize_t text_size;
    XML_Char*  text;
    int        keep_text;
    int        done;
    Py_ssize_t queue_size;
    PyObject** queue;
} IterParser;

#if BYTEORDER == 1234
#define TD_AS_INT      0x00004454
#define TD_AS_INT_MASK 0x00ffffff
#else
#define TD_AS_INT      0x54440000
#define TD_AS_INT_MASK 0xffffff00
#endif

static int
text_realloc(IterParser *self, Py_ssize_t req_size)
{
    Py_ssize_t  n       = req_size;
    char       *new_mem = NULL;

    if (req_size < self->text_alloc) {
        return 0;
    }

    /* Calculate the next-highest power of two */
    n--;
    n |= n >> 1;
    n |= n >> 2;
    n |= n >> 4;
    n |= n >> 8;
    n |= n >> 16;
    n++;

    if (n < req_size) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory");
        return -1;
    }

    new_mem = malloc(n * sizeof(XML_Char));
    if (new_mem == NULL) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory");
        return -1;
    }

    memcpy(new_mem, self->text, (self->text_size + 1) * sizeof(XML_Char));

    free(self->text);
    self->text = new_mem;
    self->text_alloc = n;

    return 0;
}

#define IS_WHITESPACE(c) ((c) == 0x20 || (c) == 0x0d || (c) == 0x0a || (c) == 0x09)

static int
text_append(IterParser *self, const XML_Char *data, Py_ssize_t len)
{
    int new_size;

    if (len == 0) {
        return 0;
    }

    /* If this is the first chunk, handle whitespace */
    if (self->text_size == 0) {
        while (len && IS_WHITESPACE(*data)) {
            ++data;
            --len;
        }
    }

    new_size = self->text_size + len;
    if (text_realloc(self, new_size + 1)) {
        return -1;
    }

    memcpy(self->text + self->text_size,
           data,
           len * sizeof(XML_Char));

    self->text_size = new_size;
    self->text[self->text_size] = 0x0;

    return 0;
}

static void
text_clear(IterParser *self)
{
    self->text[0] = 0;
    self->text_size = 0;
}

static inline PyObject*
make_pos(IterParser *self)
{
    unsigned long line;
    unsigned long col;

    line = (unsigned long)XML_GetCurrentLineNumber(self->parser);
    col = (unsigned long)XML_GetCurrentColumnNumber(self->parser);

    return Py_BuildValue("kk", line, col);
}

static void
startElement(IterParser *self, const XML_Char *name, const XML_Char **atts)
{
    PyObject*        pyname;
    PyObject*        pyatts;
    const XML_Char** att_ptr = atts;
    PyObject*        tuple;

    /* Note that we can't actually do much error checking here in this
       expat callback, since there is no way to communicate that error
       to expat. */

    /* Don't overflow the queue -- in practice this should *never* happen */
    if (self->queue_write_idx < self->queue_size) {
        tuple = PyTuple_New(4);

        Py_INCREF(Py_True);
        PyTuple_SET_ITEM(tuple, 0, Py_True);

        /* This is an egregious but effective optimization.  Since by
           far the most frequently occurring element name in a large
           VOTABLE file is TD, we explicitly check for it here with
           integer comparison to avoid the lookup in the interned
           string table in PyString_InternFromString, and return a
           singleton string for "TD" */
        if ((*(int*)name & TD_AS_INT_MASK) == TD_AS_INT) {
            Py_INCREF(self->td_singleton);
            PyTuple_SET_ITEM(tuple, 1, self->td_singleton);
        } else {
            pyname = PyString_InternFromString(name);
            PyTuple_SET_ITEM(tuple, 1, pyname);
        }

        if (*att_ptr) {
            pyatts = PyDict_New();
            do {
                PyDict_SetItem(
                    pyatts,
                    PyString_InternFromString(*att_ptr),
                    PyString_FromString(*(att_ptr + 1)));
                att_ptr += 2;
            } while (*att_ptr);
        } else {
            Py_INCREF(self->dict_singleton);
            pyatts = self->dict_singleton;
        }
        PyTuple_SET_ITEM(tuple, 2, pyatts);

        PyTuple_SET_ITEM(tuple, 3, make_pos(self));

        text_clear(self);

        self->keep_text = 1;

        self->queue[self->queue_write_idx++] = tuple;
    }
}

static void
endElement(IterParser *self, const XML_Char *name)
{
    PyObject* pyname;
    PyObject* tuple;
    PyObject* pytext;
    XML_Char* end;

    /* Note that we can't actually do much error checking here in this
       expat callback, since there is no way to communicate that error
       to expat. */

    /* Don't overflow the queue -- in practice this should *never* happen */
    if (self->queue_write_idx < self->queue_size) {
        tuple = PyTuple_New(4);

        Py_INCREF(Py_False);
        PyTuple_SET_ITEM(tuple, 0, Py_False);

        /* This is an egregious but effective optimization.  Since by
           far the most frequently occurring element name in a large
           VOTABLE file is TD, we explicitly check for it here with
           integer comparison to avoid the lookup in the interned
           string table in PyString_InternFromString, and return a
           singleton string for "TD" */
        if ((*(int*)name & TD_AS_INT_MASK) == TD_AS_INT) {
            Py_INCREF(self->td_singleton);
            PyTuple_SET_ITEM(tuple, 1, self->td_singleton);
        } else {
            pyname = PyString_InternFromString(name);
            PyTuple_SET_ITEM(tuple, 1, pyname);
        }

        /* Cut whitespace off the end of the string */
        end = self->text + self->text_size - 1;
        while (end >= self->text && IS_WHITESPACE(*end)) {
            --end;
            --self->text_size;
        }
        pytext = PyString_FromStringAndSize(self->text, self->text_size);
        PyTuple_SET_ITEM(tuple, 2, pytext);

        PyTuple_SET_ITEM(tuple, 3, make_pos(self));

        self->keep_text = 0;

        self->queue[self->queue_write_idx++] = tuple;
    }
}

static void
characterData(IterParser *self, const XML_Char *text, int len)
{
    /* Note that we can't actually do much error checking here in this
       expat callback, since there is no way to communicate that error
       to expat. */

    if (self->keep_text) {
        text_append(self, text, len);
    }
}

/* To support cyclical garbage collection */
static int
IterParser_traverse(IterParser *self, visitproc visit, void *arg)
{
    int vret;

    if (self->fd) {
        vret = visit(self->fd, arg);
        if (vret != 0) return vret;
    }

    if (self->read) {
        vret = visit(self->read, arg);
        if (vret != 0) return vret;
    }

    if (self->read_args) {
        vret = visit(self->read_args, arg);
        if (vret != 0) return vret;
    }

    if (self->dict_singleton) {
        vret = visit(self->dict_singleton, arg);
        if (vret != 0) return vret;
    }

    if (self->td_singleton) {
        vret = visit(self->td_singleton, arg);
        if (vret != 0) return vret;
    }

    return 0;
}

/* To support cyclical garbage collection */
static int
IterParser_clear(IterParser *self)
{
    PyObject *tmp;

    tmp = self->fd;
    self->fd = NULL;
    Py_XDECREF(tmp);

    tmp = self->read;
    self->read = NULL;
    Py_XDECREF(tmp);

    tmp = self->read_args;
    self->read_args = NULL;
    Py_XDECREF(tmp);

    tmp = self->dict_singleton;
    self->dict_singleton = NULL;
    Py_XDECREF(tmp);

    tmp = self->td_singleton;
    self->td_singleton = NULL;
    Py_XDECREF(tmp);

    return 0;
}

static void
IterParser_dealloc(IterParser* self)
{
    IterParser_clear(self);
    free(self->queue);
    free(self->text);
    if (self->parser != NULL) {
        XML_ParserFree(self->parser);
    }
    self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
IterParser_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    IterParser *self = NULL;

    self = (IterParser *)type->tp_alloc(type, 0);
    if (self != NULL) {
        self->parser          = NULL;
        self->fd              = NULL;
        self->read            = NULL;
        self->queue_read_idx  = 0;
        self->queue_write_idx = 0;
        self->text            = NULL;
        self->read_args       = NULL;
        self->dict_singleton  = NULL;
        self->td_singleton    = NULL;
        self->queue_size      = 0;
        self->queue           = NULL;
    }

    return (PyObject *)self;
}

static int
IterParser_init(IterParser *self, PyObject *args, PyObject *kwds)
{
    PyObject* fd         = NULL;
    int       buffersize = 1 << 14;

    static char *kwlist[] = {"fd", "buffersize", NULL};
    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|i:IterParser.__init__", kwlist,
                                     &fd, &buffersize)) {
        return -1;
    }

    self->buffersize = buffersize >= 1024 ? buffersize : 1024;
    if (PyCallable_Check(fd)) {
        self->fd = fd;   Py_INCREF(self->fd);
        self->read = fd; Py_INCREF(self->read);
    } else {
        if (!PyObject_HasAttrString(fd, "read")) {
            fd = PyFile_FromString(PyString_AsString(fd), "rb");
            if (fd == NULL) {
                return -1;
            }
        }
        self->fd = fd; Py_INCREF(self->fd);
        self->read = PyObject_GetAttrString(fd, "read");
        if (self->read == NULL) {
            goto fail;
        }
    }

    self->queue_read_idx  = 0;
    self->queue_write_idx = 0;
    self->done            = 0;

    self->text = malloc(buffersize * sizeof(XML_Char));
    self->text_alloc = buffersize;
    if (self->text == NULL) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory");
        goto fail;
    }
    text_clear(self);

    self->read_args = PyTuple_Pack(1, PyInt_FromSsize_t(buffersize));
    if (self->read_args == NULL) {
        goto fail;
    }

    self->dict_singleton = PyDict_New();
    self->td_singleton = PyString_FromString("TD");

    self->queue_size = buffersize;
    self->queue = malloc(sizeof(PyObject*) * self->queue_size);
    if (self->queue == NULL) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory");
        goto fail;
    }

    /* Set up an expat parser with our callbacks */
    self->parser = XML_ParserCreate(NULL);
    if (self->parser == NULL) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory");
        goto fail;
    }
    XML_SetUserData(self->parser, self);
    XML_SetElementHandler(
        self->parser,
        (XML_StartElementHandler)startElement,
        (XML_EndElementHandler)endElement);
    XML_SetCharacterDataHandler(
        self->parser,
        (XML_CharacterDataHandler)characterData);

    return 0;

 fail:
    Py_XDECREF(self->fd);
    Py_XDECREF(self->read);
    free(self->text);
    Py_XDECREF(self->dict_singleton);
    Py_XDECREF(self->td_singleton);
    Py_XDECREF(self->read_args);
    free(self->queue);

    return -1;
}

static PyMemberDef IterParser_members[] =
{
    {NULL}  /* Sentinel */
};

static PyObject *
IterParser_iter(IterParser* self)
{
    Py_INCREF(self);
    return (PyObject*) self;
}


static PyObject *
IterParser_next(IterParser* self)
{
    PyObject*  data = NULL;
    XML_Char*  buf;
    Py_ssize_t buflen;

    if (self->queue_read_idx < self->queue_write_idx) {
        return self->queue[self->queue_read_idx++];
    }

    if (self->done) {
        return NULL;
    }

    self->queue_read_idx = 0;
    self->queue_write_idx = 0;

    do {
        data = PyObject_CallObject(self->read, self->read_args);
        if (data == NULL) {
            return NULL;
        }

        if (PyString_AsStringAndSize(data, &buf, &buflen) == -1) {
            Py_DECREF(data);
            return NULL;
        }

        if (buflen < self->buffersize) {
            self->done = 1;
        }

        if (XML_Parse(self->parser, buf, (int)buflen, self->done) == XML_STATUS_ERROR) {
            PyErr_Format(
                PyExc_ValueError, "%lu:%lu: %s",
                XML_GetCurrentLineNumber(self->parser),
                XML_GetCurrentColumnNumber(self->parser),
                XML_ErrorString(XML_GetErrorCode(self->parser)));
            Py_DECREF(data);
            return NULL;
        }
        Py_DECREF(data);
    } while (self->queue_write_idx == 0 && self->done == 0);

    if (self->queue_write_idx == 0) {
        return NULL;
    }

    if (self->queue_write_idx >= self->queue_size) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "XML queue overflow.  This most likely indicates an internal bug.");
        return NULL;
    }

    return self->queue[self->queue_read_idx++];
}

static PyMethodDef IterParser_methods[] =
{
    {NULL}  /* Sentinel */
};

static PyTypeObject IterParserType =
{
    PyObject_HEAD_INIT(NULL)
    0,                          /*ob_size*/
    "IterParser.IterParser",    /*tp_name*/
    sizeof(IterParser),         /*tp_basicsize*/
    0,                          /*tp_itemsize*/
    (destructor)IterParser_dealloc, /*tp_dealloc*/
    0,                          /*tp_print*/
    0,                          /*tp_getattr*/
    0,                          /*tp_setattr*/
    0,                          /*tp_compare*/
    0,                          /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    0,                          /*tp_hash */
    0,                          /*tp_call*/
    0,                          /*tp_str*/
    0,                          /*tp_getattro*/
    0,                          /*tp_setattro*/
    0,                          /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
    "IterParser objects",       /* tp_doc */
    (traverseproc)IterParser_traverse, /* tp_traverse */
    (inquiry)IterParser_clear,  /* tp_clear */
    0,                          /* tp_richcompare */
    0,                          /* tp_weaklistoffset */
    (getiterfunc)IterParser_iter, /* tp_iter */
    (iternextfunc)IterParser_next, /* tp_iternext */
    IterParser_methods,         /* tp_methods */
    IterParser_members,         /* tp_members */
    0,                          /* tp_getset */
    0,                          /* tp_base */
    0,                          /* tp_dict */
    0,                          /* tp_descr_get */
    0,                          /* tp_descr_set */
    0,                          /* tp_dictoffset */
    (initproc)IterParser_init,  /* tp_init */
    0,                          /* tp_alloc */
    IterParser_new,             /* tp_new */
};

static PyMethodDef module_methods[] =
{
    {NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
inititerparser(void)
{
    PyObject* m;

    if (PyType_Ready(&IterParserType) < 0)
        return;

    m = Py_InitModule3("iterparser", module_methods, "Fast XML parser");

    if (m == NULL)
      return;

    Py_INCREF(&IterParserType);
    PyModule_AddObject(m, "IterParser", (PyObject *)&IterParserType);
}
