# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Exceptions and warnings used by the vo package.
"""

from __future__ import division

from warnings import warn

class VOTableSpecWarning(SyntaxWarning):
    """
    Warning emitted when the input XML file violates the spec, but
    there is an obvious workaround.
    """
    pass

class UnimplementedWarning(SyntaxWarning):
    """
    Warning emitted when the a feature of VOTABLE is not implemented.
    """
    pass

class VOTableSpecError(ValueError):
    """
    Error raised when the input XML file violates the spec and there
    is no good workaround.
    """
    pass

def _format_message(message, config={}, pos=None):
    if pos is None:
        pos = ('?', '?')
    filename = config.get('filename', '?')
    return '%s:%s:%s: %s' % (filename, pos[0], pos[1], message)

def warn_or_raise(message, config={}, pos=None, exc=VOTableSpecError,
                  wrn=VOTableSpecWarning):
    """
    Warn or raise an exception, depending on the pedantic setting.
    """
    message = _format_message(message, config, pos)
    if config.get('pedantic'):
        raise exc(message)
    else:
        warn(message, wrn)

def vo_raise(message, config={}, pos=None, exc=VOTableSpecError):
    """
    Raise an exception, with proper position information if available.
    """
    message = _format_message(message, config, pos)
    raise exc(message)

def vo_reraise(exc, config={}, pos=None):
    """
    Raise an exception, with proper position information if available.
    """
    message = _format_message(str(exc), config, pos)
    exc.message = message
    raise exc

def vo_warn(message, config={}, pos=None, wrn=VOTableSpecWarning):
    """
    Warn, with proper position information if available.
    """
    message = _format_message(message, config, pos)
    warn(message, wrn)
